'use strict';

var _interopRequireDefault = function (obj) { return obj && obj.__esModule ? obj : { 'default': obj }; };

var _classCallCheck = function (instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError('Cannot call a class as a function'); } };

var _createClass = (function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ('value' in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; })();

var _get = function get(object, property, receiver) { var desc = Object.getOwnPropertyDescriptor(object, property); if (desc === undefined) { var parent = Object.getPrototypeOf(object); if (parent === null) { return undefined; } else { return get(parent, property, receiver); } } else if ('value' in desc) { return desc.value; } else { var getter = desc.get; if (getter === undefined) { return undefined; } return getter.call(receiver); } };

var _inherits = function (subClass, superClass) { if (typeof superClass !== 'function' && superClass !== null) { throw new TypeError('Super expression must either be null or a function, not ' + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) subClass.__proto__ = superClass; };

Object.defineProperty(exports, '__esModule', {
    value: true
});

var _assign = require('object-assign');

var _assign2 = _interopRequireDefault(_assign);

var _EventEmitter2 = require('eventemitter3');

var _EventEmitter3 = _interopRequireDefault(_EventEmitter2);

var _toposort = require('toposort');

var _toposort2 = _interopRequireDefault(_toposort);

var _getAllPropertyNames = require('getallpropertynames');

var _getAllPropertyNames2 = _interopRequireDefault(_getAllPropertyNames);

var _extend = require('./util/extend');

var _extend2 = _interopRequireDefault(_extend);

var _hasPrefix = require('./util/hasPrefix');

var _hasPrefix2 = _interopRequireDefault(_hasPrefix);

var _Actions = require('./Actions');

var _Actions2 = _interopRequireDefault(_Actions);

var _Store = require('./Store');

var _Store2 = _interopRequireDefault(_Store);

var allObjectProperties = _getAllPropertyNames2['default']({});
var allActionsProperties = _getAllPropertyNames2['default'](_Actions2['default'].prototype);
var eventEmitterProperties = Object.keys(_EventEmitter3['default'].prototype);

var Dispatcher = (function (_EventEmitter) {

    /**
     * Constructor
     * @param  {Object} options
     * @param  {Object} options.actions Namespaced actions
     * @param  {Object} options.stores  Namespaced stores
     * @return {Dispatcher}
     */

    function Dispatcher(options) {
        _classCallCheck(this, Dispatcher);

        _get(Object.getPrototypeOf(Dispatcher.prototype), 'constructor', this).call(this);
        // Wrapped actions
        this.actions = {};
        this.actionIds = [];
        // Wrapped stores
        this.stores = {};
        // Actual stores
        this._stores = {};
        // Order in which actions get dispatched to the stores
        this.order = [];

        var actions = options.actions || {};
        var stores = options.stores || {};

        this.createActions(actions);
        this.createStores(stores);
    }

    _inherits(Dispatcher, _EventEmitter);

    _createClass(Dispatcher, [{
        key: 'dispatch',

        /**
         * Dispatches an action
         * @param  {String}    id   The id of the action (e.g. 'todos.create')
         * @param  {...mixed}  args Arguments that will be passed to the handlers
         * @return {void}
         */
        value: function dispatch(id) {
            for (var _len = arguments.length, args = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
                args[_key - 1] = arguments[_key];
            }

            if (this._isDispatching) {
                throw new Error('Cannot dispatch in the middle of a dispatch.');
            }

            this._isDispatching = true;

            this.emit.apply(this, ['dispatch', id].concat(args));

            // Run through stores and invoke registered handlers
            var stores = this._stores;
            for (var i = 0; i < this.order.length; i++) {
                var key = this.order[i];
                var handlers = stores[key]._handlers;
                if (!handlers || !handlers[id]) continue;

                try {
                    handlers[id].apply(handlers, args);
                } catch (err) {
                    this.emit('error', err);
                    throw err;
                }
            }

            this._isDispatching = false;
        }
    }, {
        key: 'createActions',

        /**
         * Create actions
         * @param  {Object} actions Namespaced actions
         * @return {void}
         */
        value: function createActions(actions) {
            var _this = this;

            var _loop = function (key) {
                var Actions = actions[key];
                // Make actions available at construction time
                var ExtendedActions = _extend2['default'](Actions, {
                    key: key,
                    actions: _this.actions,
                    stores: _this.stores
                });

                // Instantiate actions
                var instance = new ExtendedActions();
                // Create wrapped actions object
                _this.actions[key] = {};
                // Find actual action function
                var props = _getAllPropertyNames2['default'](Actions.prototype).filter(function (prop) {
                    // Ignore the base class properties
                    return allActionsProperties.indexOf(prop) < 0 &&
                    // Only regard functions
                    typeof instance[prop] == 'function';
                });
                // Run through actual actions
                for (var i = 0; i < props.length; i++) {
                    var prop = props[i];
                    // Bind function to instance
                    var fn = instance[prop] = instance[prop].bind(instance);
                    // The action id is composed from the actions key and its function name
                    var id = [key, prop].join('.');
                    _this.actionIds.push(id);
                    // Listen to the action event
                    instance.addListener(prop, _this.dispatch.bind(_this, id));
                    // Add function to the wrapped object
                    _this.actions[key][prop] = fn;
                }
            };

            // Run through namespaced actions
            for (var key in actions) {
                _loop(key);
            }
        }
    }, {
        key: 'actionIdExists',

        /**
         * Check if action id exists
         * @param  {String} id The action id
         * @return {Boolean}
         */
        value: function actionIdExists(id) {
            return this.actionIds.indexOf(id) > -1;
        }
    }, {
        key: 'createStores',

        /**
         * Create stores
         * @param  {Object} stores Namespaced stores
         * @return {void}
         */
        value: function createStores(stores) {
            var _this2 = this;

            var nodes = [];
            var edges = [];

            // Create a dependency graph
            for (var key in stores) {
                nodes.push(key);
                var store = stores[key];
                // If store is not an array, it has no dependencies
                if (!Array.isArray(store)) continue;
                var deps = store.slice(1) || [];
                // Add edges between store and it dependencies
                for (var i = 0; i < deps.length; i++) {
                    edges.push([key, deps[i]]);
                }
            }

            // Topological sort, store the order
            var order = this.order = _toposort2['default'].array(nodes, edges).reverse();

            var _loop2 = function (i, l) {
                var key = order[i];
                var Store = stores[key];
                // Handle plain and array definition
                if (Array.isArray(Store)) Store = Store[0];
                // Make stores available at construction time
                var ExtendedStore = _extend2['default'](Store, {
                    key: key,
                    stores: _this2.stores,
                    _actionIdExists: _this2.actionIdExists.bind(_this2)
                });
                // Instantiate the store
                var instance = new ExtendedStore();
                _this2._stores[key] = instance;
                // Create a wrapped stores object
                _this2.stores[key] = {};
                // Find functions that will be added to the wrapped object
                var props = _getAllPropertyNames2['default'](Store.prototype).filter(function (prop) {
                    // Only regard functions
                    return typeof instance[prop] === 'function' && allObjectProperties.indexOf(prop) < 0 && (_hasPrefix2['default'](prop, 'get') || _hasPrefix2['default'](prop, 'is') || _hasPrefix2['default'](prop, 'has') || eventEmitterProperties.indexOf(prop) > -1 && prop !== 'emit');
                });
                // Run through functions
                for (var _i = 0; _i < props.length; _i++) {
                    var prop = props[_i];
                    // Bind function to the instance and add it to the wrapped object
                    _this2.stores[key][prop] = instance[prop].bind(instance);
                }
            };

            // Run through ordered stores
            for (var i = 0, l = order.length; i < l; i++) {
                _loop2(i, l);
            }
        }
    }]);

    return Dispatcher;
})(_EventEmitter3['default']);

exports['default'] = Dispatcher;
module.exports = exports['default'];

// Functions that start with get or is

// Event emitter function, except emit